window.addEventListener("load", function () {
  dictionarySwitcher.init();
}, false);

var dictionarySwitcher = {
  currentDict: "",
  dictMenu: {},
  panel: null,
  rememberMenuitem: null,
  list: [],
  langToDict: {},
  tldToDict: {},
  countryToTld: {
    de: ['at', 'ch'],
    at: ['de', 'ch'],
    ch: ['de', 'at'],
    us: ['uk', 'com', 'org', 'net', 'gov', 'edu', 'au'],
    gb: ['uk', 'com', 'org', 'net', 'gov', 'edu', 'au'],
    au: ['uk', 'com', 'org', 'net', 'gov', 'edu']
  },
  autoSucceeded: false,
  init: function () {
    if (!gURLBar)
      gURLBar = document.getElementById("urlbar");
    this.panel = document.getElementById("dictionary-switcher");
    this.l10n = document.getElementById("dictionary-switcher-l10n");
    this.rememberMenuitem = document.getElementById("dictionary-switcher-remember");
    this.menuseparator = document.getElementById("dictionary-switcher-menu").getElementsByTagName("menuseparator")[0];

    this.prefs = Cc["@mozilla.org/preferences-service;1"]
                   .getService(Ci.nsIPrefService)
                   .getBranch("spellchecker.")
                   .QueryInterface(Ci.nsIPrefBranch2);
    try {
      this.currentDict = this.prefs.getCharPref("dictionary");
    } catch (e) {}
    this.prefs.addObserver("", this, false);
    
    let (spellchecker = Cc["@mozilla.org/spellchecker/" + ("@mozilla.org/spellchecker/myspell;1" in Cc ? "myspell;1" : "engine;1")]
                          .createInstance(Ci.mozISpellCheckingEngine))
      spellchecker.getDictionaryList(this.list, {});
    this.list = this.list.value.toString().split(",");
    
    let (menuitem = document.createElement("menuitem"),
         bundleService = Cc["@mozilla.org/intl/stringbundle;1"].getService(Ci.nsIStringBundleService)) {
      let languageBundle = bundleService.createBundle("chrome://global/locale/languageNames.properties");
      let regionBundle = bundleService.createBundle("chrome://global/locale/regionNames.properties");
      menuitem.setAttribute("type", "radio");
      menuitem.setAttribute("name", "dictionary-switcher-dicts");
      menuitem.setAttribute("oncommand", "dictionarySwitcher.toggle(this.getAttribute('dict'))");
      for (let i=0; i < this.list.length; i++) {
        let dict = this.list[i];
        let code = dict.split(/-|_/);
        this.langToDict[code[0]] = dict;
        if (code[1]) {
          let id = code[1].toLowerCase();
          this.tldToDict[id] = dict;
          if (id in this.countryToTld) {
            this.countryToTld[id].forEach(function (tld) {
              if (!(tld in this.tldToDict))
                this.tldToDict[tld] = dict;
            }, this);
          }
        }
        
        let displayName;
        let isoStrArray = dict.split(/-|_/);
        if (isoStrArray[0]) {
          try {
            displayName = languageBundle.GetStringFromName(isoStrArray[0].toLowerCase());
          } catch (e) {}
          if (isoStrArray[1]) {
            try {
              displayName += " / " + regionBundle.GetStringFromName(isoStrArray[1].toLowerCase());
            } catch (e) {}
            if (isoStrArray[2])
              displayName += " (" + isoStrArray[2] + ")";
          }
        }

        menuitem = menuitem.cloneNode(true);
        menuitem.setAttribute("label", displayName || dict);
        menuitem.setAttribute("dict", dict);
        this.dictMenu[dict] = menuitem;
        this.menuseparator.parentNode.insertBefore(menuitem, this.menuseparator);
      }
    }

    let (obj = this) {
      gURLBar.addEventListener("ValueChange", function () {
        //XXX bug 436717
        setTimeout(function () {
          obj.pageSwitch();
        }, 0);
      }, true);
      this._pageLoad = function (e) {
        obj.pageLoad(e);
      };
      this._keyPress = function (e) {
        obj.keyPress(e);
      };
      window.addEventListener("unload", function () {
        obj.prefs = null;
        obj._textarea = null;
      }, false);
    }
    this.getAutoPref();
    this.getAsITypePref();

    this.hostToDict = Components.utils.evalInSandbox(
      this.prefs.getCharPref("hostToDict"),
      new Components.utils.Sandbox("about:blank"));
    this.updateUI();
  },
  observe: function (subject, topic, data) {
    if (topic == "nsPref:changed") {
      switch (data) {
        case "dictionary":
          try {
            this.currentDict = this.prefs.getCharPref("dictionary");
          } catch(e) {
            this.currentDict = null;
          }
          this.updateUI();
          break;
        case "autoChangeDictionary":
          this.getAutoPref();
          if (this.auto)
            this.autoDetect(1,1);
          this.updateUI();
          break;
        case "detectDictionaryAsIType":
          this.getAsITypePref();
          break;
      }
    }
  },
  getAutoPref: function () {
    this.auto = this.prefs.getBoolPref("autoChangeDictionary");
    if (this.auto) {
      document.getElementById("dictionary-switcher-auto").setAttribute("checked", "true");
      getBrowser().addEventListener("DOMContentLoaded", this._pageLoad, true);
    } else {
      document.getElementById("dictionary-switcher-auto").removeAttribute("checked");
      getBrowser().removeEventListener("DOMContentLoaded", this._pageLoad, true);
    }
  },
  getAsITypePref: function () {
    this.asIType = this.prefs.getBoolPref("detectDictionaryAsIType");
    if (this.asIType) {
      document.getElementById("dictionary-switcher-as-i-type").setAttribute ("checked", "true");
      getBrowser().addEventListener("keypress", this._keyPress, true);
    } else {
      document.getElementById("dictionary-switcher-as-i-type").removeAttribute ("checked");
      getBrowser().removeEventListener("keypress", this._keyPress, true);
    }
  },
  canSpellCheck: function (node) {
    return node &&
           (node instanceof Window || node instanceof HTMLHtmlElement ||
            (node instanceof HTMLTextAreaElement || (node instanceof HTMLInputElement && node.type != "password")) && !node.readOnly);
  },
  getEditor: function (node) {
    if (node instanceof HTMLHtmlElement)
      node = node.ownerDocument.defaultView;
    if (node instanceof Window) {
      let session = node.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIWebNavigation).QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIEditingSession);
      if (session.windowIsEditable(node))
        return session.getEditorForWindow(node);
    } else if (!node.readOnly) {
      return node.QueryInterface(Ci.nsIDOMNSEditableElement).editor;
    }
    return null;
  },
  keyPress: function (e) {
    var node = e.originalTarget;
    if ((this.host && this.host in this.hostToDict) || !this.canSpellCheck(node))
      return;
    var text = node instanceof HTMLHtmlElement ? node.ownerDocument.body.textContent : node.value;
    if (text.length == 10) {
      text = text.replace(/\W+/g, " ").replace(/^ /,"").replace(/ $/,"").split(" ");
      if (text.length < 3)
        return;
    } else if (text.length > 15 && node != this._textarea) {
      text = text.replace(/\W+/g, " ").replace(/^ /,"").substring(0, 15).replace(/ $/,"").split(" ");
    } else {
      return;
    }
    this._textarea = node;
    var editor = this.getEditor(node);
    if (!editor)
      return;
    var sc = editor.getInlineSpellChecker ?
      editor.getInlineSpellChecker(true) :
      editor.inlineSpellChecker;
    if (!sc || !sc.spellChecker)
      return;
    var maxMisspelled = Math.round(text.length / 4);
    var initialDict = sc.spellChecker.GetCurrentDictionary(), dict = initialDict;
    var misspelled, oldMisspelled, newDict;
    do {
      misspelled = 0;
      if (dict != initialDict)
        sc.spellChecker.SetCurrentDictionary(dict);
      for (let i = 0; i < text.length; i++)
        if (sc.spellChecker.CheckCurrentWord(text[i]))
          misspelled++;
      if (!newDict || misspelled < oldMisspelled)
        newDict = dict;
      oldMisspelled = misspelled;
      dict = this.getNextInList(dict);
    } while (misspelled > maxMisspelled && dict != initialDict);
    if (newDict && newDict != initialDict)
      this.toggle(newDict, true);
    else
      sc.spellChecker.SetCurrentDictionary(initialDict);
  },
  tryRecheck: function () {
    var fields = [document.commandDispatcher.focusedElement || document.commandDispatcher.focusedWindow]; // if a field is focused, recheck it
    if (!this.canSpellCheck(fields[0]))
      fields = getBrowser().selectedBrowser.docShell.QueryInterface(Ci.nsIWebNavigation).document.getElementsByTagName("textarea"); // otherwise recheck all textareas
    for (let i = 0; i < fields.length; i++) {
      let editor = this.getEditor(fields[i]);
      if (editor) {
        let sc = editor.getInlineSpellChecker ?
          editor.getInlineSpellChecker(true) :
          editor.inlineSpellChecker;
        if (sc && sc.spellChecker) {
          sc.spellChecker.SetCurrentDictionary(this.currentDict);
          sc.spellCheckRange(null);
        }
      }
    }
  },
  setDict: function (dict, recheck) {
    if (!dict)
      return false;
    if (this.currentDict != dict) {
      this.prefs.setCharPref("dictionary", this.currentDict = dict);
      if (recheck)
        this.tryRecheck();
    }
    return true;
  },
  pageLoad: function (e) {
    this._textarea = null;
    if (e.originalTarget == getBrowser().selectedBrowser.docShell.QueryInterface(Ci.nsIWebNavigation).document)
      this.autoDetect(0,1);
  },
  pageSwitch: function () {
    this._textarea = null;
    this.getHost();
    var dict = this.hostToDict[this.host];
    if (dict)
      this.setDict(dict);
    else
      this.autoDetect(1,1);
    this.updateUI();
  },
  autoDetect: function (checkHost, checkDocLang) {
    if (!this.auto)
      return;
    if (this.host && this.host in this.hostToDict)
      return;
    if (checkDocLang)
      this.autoSucceeded = this.autoDetectPerDocLang();
    if (this.host && checkHost && !this.autoSucceeded)
      this.autoSucceeded = this.autoDetectPerHost();
  },
  autoDetectPerHost: function () {
    if (!this.host)
      return false;
    var tld = this.host.replace(/.*\./, "");
    return tld in this.tldToDict ?
           this.setDict(this.tldToDict[tld]) :
           false;
  },
  autoDetectPerDocLang: function () {
    var root = getBrowser().selectedBrowser.docShell.QueryInterface(Ci.nsIWebNavigation).document.documentElement;
    if (root) {
      let lang = root.getAttribute("lang") || root.getAttribute("xml:lang");
      if (lang) {
        let i = this.list.indexOf(lang);
        return this.setDict(i < 0 ?
          this.langToDict[lang.substr(0,2)] :
          this.list[i], true);
      }
    }
    return false;
  },
  updateUI: function () {
    var saved = this.host in this.hostToDict;
    this.panel.label = (this.auto && this.autoSucceeded && !saved ? this.l10n.getString("autoDetected")+" " : "") + (this.currentDict || this.l10n.getString("noDict"));
    this.panel.setAttribute("saved", saved);
    this.rememberMenuitem.setAttribute("checked", saved);
    if (this.dictMenu._last)
      this.dictMenu._last.removeAttribute("checked");
    this.dictMenu._last = this.dictMenu[ this.currentDict ];
    if (this.dictMenu._last)
      this.dictMenu._last.setAttribute("checked", true);
  },
  getNextInList: function (dict) {
    var i = this.list.indexOf(dict) + 1;
    if (i >= this.list.length)
      i = 0;
    return this.list[i];
  },
  toggle: function (dict, dontSave) {
    this.setDict(dict || this.getNextInList(this.currentDict), true);
    if (this.host && !dontSave) {
      this.hostToDict[this.host] = this.currentDict;
      this.syncHostList();
    }
  },
  toggleAuto: function () {
    this.prefs.setBoolPref("autoChangeDictionary", !this.auto);
  },
  toggleAsIType: function () {
    this.prefs.setBoolPref("detectDictionaryAsIType", !this.asIType);
  },
  toggleRemember: function () {
    if (this.host) {
      if (this.rememberMenuitem.getAttribute("checked") == "true")
        this.hostToDict[this.host] = this.currentDict;
      else {
        delete this.hostToDict[this.host];
        this.autoDetect(1,1);
      }
      this.syncHostList();
    }
  },
  syncHostList: function () {
    this.prefs.setCharPref("hostToDict", this.hostToDict.toSource());
    this.updateUI();
  },
  getHost: function () {
    var url = gURLBar.value.match(/\/\/(.*?)\//);
    this.host = url ? url[1] : null;
  }
};
